﻿using System;

namespace SpatialAHRS_Example {

	public class ComplementaryAHRS {

		const double kGravity = 1.00;
		const double gamma_ = 0.01;
		// Bias estimation steady state thresholds
		const double kAngularVelocityThreshold = 0.2;
		const double kAccelerationThreshold = 0.1;
		const double kDeltaAngularVelocityThreshold = 0.01;

		// Gain parameter for the complementary filter, belongs in [0, 1].
		double gain_acc_;
		public double gain_acc {
			set {
				if (value >= 0 && value <= 1.0) {
					gain_acc_ = value;
				}
			}
			get {
				return gain_acc_;
			}
		}
		double gain_mag_;
		public double gain_mag {
			set {
				if (value >= 0 && value <= 1.0) {
					gain_mag_ = value;
				}
			}
			get {
				return gain_mag_;
			}
		}

		// Bias estimation gain parameter, belongs in [0, 1].
		double bias_alpha_;
		public double bias_alpha {
			set {
				if (value >= 0 && value <= 1.0) {
					bias_alpha_ = value;
				}
			}
			get {
				return bias_alpha_;
			}
		}

		// Parameter whether to do bias estimation or not.
		public bool do_bias_estimation { get; set; }

		// Parameter whether to do adaptive gain or not.
		public bool do_adaptive_gain_ { get; set; }

		bool initialized_;
		public bool steady_state { get; set; }

		// The orientation as a Hamilton quaternion (q0 is the scalar). Represents
		// the orientation of the fixed frame wrt the body frame.
		double q0_, q1_, q2_, q3_;

		// Bias in angular velocities;
		double wx_prev_, wy_prev_, wz_prev_;

		// Bias in angular velocities;
		double wx_bias_, wy_bias_, wz_bias_;

		public ComplementaryAHRS() {
			gain_acc_ = 0.01;
			gain_mag_ = 0.005;
			bias_alpha_ = 0.01;
			do_bias_estimation = true;
			do_adaptive_gain_ = true;
			initialized_ = false;
			steady_state = false;
			q0_ = q1_ = q2_ = q3_ = 0;
			wx_prev_ = wy_prev_ = wz_prev_ = 0;
			wx_bias_ = wy_bias_ = wz_bias_ = 0;
		}

		public void setOrientation(
			double q0, double q1, double q2, double q3) {
			// Set the state to inverse (state is fixed wrt body).
			invertQuaternion(q0, q1, q2, q3, ref q0_, ref q1_, ref q2_, ref q3_);
		}


		public void update(double ax, double ay, double az,
										 double wx, double wy, double wz,
										 double dt) {
			// This will mess up the algorithm
			if (ax == 0 && ay == 0 && az == 0)
				return;

			if (!initialized_) {
				// First time - ignore prediction:
				getMeasurement(ax, ay, az,
							   ref q0_, ref q1_, ref q2_, ref q3_);
				initialized_ = true;
				return;
			}

			// Bias estimation.
			if (do_bias_estimation)
				updateBiases(ax, ay, az, wx, wy, wz);

			// Prediction.
			double q0_pred = 0, q1_pred = 0, q2_pred = 0, q3_pred = 0;
			getPrediction(wx, wy, wz, dt,
						  ref q0_pred, ref q1_pred, ref q2_pred, ref q3_pred);

			// Correction (from acc):
			// q_ = q_pred * [(1-gain) * qI + gain * dq_acc]
			// where qI = identity quaternion
			double dq0_acc = 0, dq1_acc = 0, dq2_acc = 0, dq3_acc = 0;
			getAccCorrection(ax, ay, az,
							 q0_pred, q1_pred, q2_pred, q3_pred,
							 ref dq0_acc, ref dq1_acc, ref dq2_acc, ref dq3_acc);

			double gain;
			if (do_adaptive_gain_) {
				gain = getAdaptiveGain(gain_acc_, ax, ay, az);

			} else {
				gain = gain_acc_;

			}

			scaleQuaternion(gain, ref dq0_acc, ref dq1_acc, ref dq2_acc, ref dq3_acc);

			quaternionMultiplication(q0_pred, q1_pred, q2_pred, q3_pred,
									 dq0_acc, dq1_acc, dq2_acc, dq3_acc,
									 ref q0_, ref q1_, ref q2_, ref q3_);

			normalizeQuaternion(ref q0_, ref q1_, ref q2_, ref q3_);
		}

		public void update(double ax, double ay, double az,
										 double wx, double wy, double wz,
										 double mx, double my, double mz,
										 double dt) {
			// These will mess up the algorithm
			if (mx == 0 && my == 0 && mz == 0)
				return;
			if (ax == 0 && ay == 0 && az == 0)
				return;

			if (!initialized_) {
				// First time - ignore prediction:
				getMeasurement(ax, ay, az,
							   mx, my, mz,
							   ref q0_, ref q1_, ref q2_, ref q3_);
				initialized_ = true;
				return;
			}

			// Bias estimation.
			if (do_bias_estimation)
				updateBiases(ax, ay, az, wx, wy, wz);

			// Prediction.
			double q0_pred = 0, q1_pred = 0, q2_pred = 0, q3_pred = 0;
			getPrediction(wx, wy, wz, dt,
						  ref q0_pred, ref q1_pred, ref q2_pred, ref q3_pred);

			// Correction (from acc):
			// q_temp = q_pred * [(1-gain) * qI + gain * dq_acc]
			// where qI = identity quaternion
			double dq0_acc = 0, dq1_acc = 0, dq2_acc = 0, dq3_acc = 0;
			getAccCorrection(ax, ay, az,
							 q0_pred, q1_pred, q2_pred, q3_pred,
							 ref dq0_acc, ref dq1_acc, ref dq2_acc, ref dq3_acc);
			double alpha = gain_acc_;
			if (do_adaptive_gain_)
				alpha = getAdaptiveGain(gain_acc_, ax, ay, az);
			scaleQuaternion(alpha, ref dq0_acc, ref dq1_acc, ref dq2_acc, ref dq3_acc);

			double q0_temp = 0, q1_temp = 0, q2_temp = 0, q3_temp = 0;
			quaternionMultiplication(q0_pred, q1_pred, q2_pred, q3_pred,
									 dq0_acc, dq1_acc, dq2_acc, dq3_acc,
									 ref q0_temp, ref q1_temp, ref q2_temp, ref q3_temp);

			// Correction (from mag):
			// q_ = q_temp * [(1-gain) * qI + gain * dq_mag]
			// where qI = identity quaternion
			double dq0_mag = 0, dq1_mag = 0, dq2_mag = 0, dq3_mag = 0;
			getMagCorrection(mx, my, mz,
							 q0_temp, q1_temp, q2_temp, q3_temp,
							 ref dq0_mag, ref dq1_mag, ref dq2_mag, ref dq3_mag);

			scaleQuaternion(gain_mag_, ref dq0_mag, ref dq1_mag, ref dq2_mag, ref dq3_mag);

			quaternionMultiplication(q0_temp, q1_temp, q2_temp, q3_temp,
									 dq0_mag, dq1_mag, dq2_mag, dq3_mag,
									 ref q0_, ref q1_, ref q2_, ref q3_);

			normalizeQuaternion(ref q0_, ref q1_, ref q2_, ref q3_);
		}

		bool checkState(double ax, double ay, double az,
											 double wx, double wy, double wz) {
			double acc_magnitude = Math.Sqrt(ax * ax + ay * ay + az * az);
			if (Math.Abs(acc_magnitude - kGravity) > kAccelerationThreshold)
				return false;

			if (Math.Abs(wx - wx_prev_) > kDeltaAngularVelocityThreshold ||
				Math.Abs(wy - wy_prev_) > kDeltaAngularVelocityThreshold ||
				Math.Abs(wz - wz_prev_) > kDeltaAngularVelocityThreshold)
				return false;

			if (Math.Abs(wx - wx_bias_) > kAngularVelocityThreshold ||
				Math.Abs(wy - wy_bias_) > kAngularVelocityThreshold ||
				Math.Abs(wz - wz_bias_) > kAngularVelocityThreshold)
				return false;

			return true;
		}

		void updateBiases(double ax, double ay, double az,
											   double wx, double wy, double wz) {
			steady_state = checkState(ax, ay, az, wx, wy, wz);

			if (steady_state) {
				wx_bias_ += bias_alpha_ * (wx - wx_bias_);
				wy_bias_ += bias_alpha_ * (wy - wy_bias_);
				wz_bias_ += bias_alpha_ * (wz - wz_bias_);
			}

			wx_prev_ = wx;
			wy_prev_ = wy;
			wz_prev_ = wz;
		}

		void getPrediction(
			double wx, double wy, double wz, double dt,
			ref double q0_pred, ref double q1_pred, ref double q2_pred, ref double q3_pred) {
			double wx_unb = wx - wx_bias_;
			double wy_unb = wy - wy_bias_;
			double wz_unb = wz - wz_bias_;

			q0_pred = q0_ + 0.5 * dt * (wx_unb * q1_ + wy_unb * q2_ + wz_unb * q3_);
			q1_pred = q1_ + 0.5 * dt * (-wx_unb * q0_ - wy_unb * q3_ + wz_unb * q2_);
			q2_pred = q2_ + 0.5 * dt * (wx_unb * q3_ - wy_unb * q0_ - wz_unb * q1_);
			q3_pred = q3_ + 0.5 * dt * (-wx_unb * q2_ + wy_unb * q1_ - wz_unb * q0_);

			normalizeQuaternion(ref q0_pred, ref q1_pred, ref q2_pred, ref q3_pred);
		}

		void getMeasurement(
			double ax, double ay, double az,
			double mx, double my, double mz,
			ref double q0_meas, ref double q1_meas, ref double q2_meas, ref double q3_meas) {
			// q_acc is the quaternion obtained from the acceleration vector representing
			// the orientation of the Global frame wrt the Local frame with arbitrary yaw
			// (intermediary frame). q3_acc is defined as 0.
			double q0_acc, q1_acc, q2_acc, q3_acc;

			// Normalize acceleration vector.
			normalizeVector(ref ax, ref ay, ref az);
			if (az >= 0) {
				q0_acc = Math.Sqrt((az + 1) * 0.5);
				q1_acc = -ay / (2.0 * q0_acc);
				q2_acc = ax / (2.0 * q0_acc);
				q3_acc = 0;
			} else {
				double X = Math.Sqrt((1 - az) * 0.5);
				q0_acc = -ay / (2.0 * X);
				q1_acc = X;
				q2_acc = 0;
				q3_acc = ax / (2.0 * X);
			}

			// [lx, ly, lz] is the magnetic field reading, rotated into the intermediary
			// frame by the inverse of q_acc.
			// l = R(q_acc)^-1 m
			double lx = (q0_acc * q0_acc + q1_acc * q1_acc - q2_acc * q2_acc) * mx +
				2.0 * (q1_acc * q2_acc) * my - 2.0 * (q0_acc * q2_acc) * mz;
			double ly = 2.0 * (q1_acc * q2_acc) * mx + (q0_acc * q0_acc - q1_acc * q1_acc +
				q2_acc * q2_acc) * my + 2.0 * (q0_acc * q1_acc) * mz;

			// q_mag is the quaternion that rotates the Global frame (North West Up) into
			// the intermediary frame. q1_mag and q2_mag are defined as 0.
			double gamma = lx * lx + ly * ly;
			double beta = Math.Sqrt(gamma + lx * Math.Sqrt(gamma));
			double q0_mag = beta / (Math.Sqrt(2.0 * gamma));
			double q3_mag = ly / (Math.Sqrt(2.0) * beta);

			// The quaternion multiplication between q_acc and q_mag represents the
			// quaternion, orientation of the Global frame wrt the local frame.
			// q = q_acc times q_mag
			quaternionMultiplication(q0_acc, q1_acc, q2_acc, q3_acc,
									 q0_mag, 0, 0, q3_mag,
									 ref q0_meas, ref q1_meas, ref q2_meas, ref q3_meas);
			//q0_meas = q0_acc*q0_mag;
			//q1_meas = q1_acc*q0_mag + q2_acc*q3_mag;
			//q2_meas = q2_acc*q0_mag - q1_acc*q3_mag;
			//q3_meas = q0_acc*q3_mag;
		}


		void getMeasurement(
			double ax, double ay, double az,
			ref double q0_meas, ref double q1_meas, ref double q2_meas, ref double q3_meas) {
			// q_acc is the quaternion obtained from the acceleration vector representing
			// the orientation of the Global frame wrt the Local frame with arbitrary yaw
			// (intermediary frame). q3_acc is defined as 0.

			// Normalize acceleration vector.
			normalizeVector(ref ax, ref ay, ref az);

			if (az >= 0) {
				q0_meas = Math.Sqrt((az + 1) * 0.5);
				q1_meas = -ay / (2.0 * q0_meas);
				q2_meas = ax / (2.0 * q0_meas);
				q3_meas = 0;
			} else {
				double X = Math.Sqrt((1 - az) * 0.5);
				q0_meas = -ay / (2.0 * X);
				q1_meas = X;
				q2_meas = 0;
				q3_meas = ax / (2.0 * X);
			}
		}

		void getAccCorrection(
		  double ax, double ay, double az,
		  double p0, double p1, double p2, double p3,
		  ref double dq0, ref double dq1, ref double dq2, ref double dq3) {
			// Normalize acceleration vector.
			normalizeVector(ref ax, ref ay, ref az);

			// Acceleration reading rotated into the world frame by the inverse predicted
			// quaternion (predicted gravity):
			double gx = 0, gy = 0, gz = 0;
			rotateVectorByQuaternion(ax, ay, az,
									 p0, -p1, -p2, -p3,
									 ref gx, ref gy, ref gz);

			// Delta quaternion that rotates the predicted gravity into the real gravity:
			dq0 = Math.Sqrt((gz + 1) * 0.5);
			dq1 = -gy / (2.0 * dq0);
			dq2 = gx / (2.0 * dq0);
			dq3 = 0.0;
		}

		void getMagCorrection(
		  double mx, double my, double mz,
		  double p0, double p1, double p2, double p3,
		  ref double dq0, ref double dq1, ref double dq2, ref double dq3) {
			// Magnetic reading rotated into the world frame by the inverse predicted
			// quaternion:
			double lx = 0, ly = 0, lz = 0;
			rotateVectorByQuaternion(mx, my, mz,
									 p0, -p1, -p2, -p3,
									 ref lx, ref ly, ref lz);

			// Delta quaternion that rotates the l so that it lies in the xz-plane
			// (points north):
			double gamma = lx * lx + ly * ly;
			double beta = Math.Sqrt(gamma + lx * Math.Sqrt(gamma));
			dq0 = beta / (Math.Sqrt(2.0 * gamma));
			dq1 = 0.0;
			dq2 = 0.0;
			dq3 = ly / (Math.Sqrt(2.0) * beta);
		}

		public double[] getOrientation() {

			// Return the inverse of the state (state is fixed wrt body).
			double q0 = 0, q1 = 0, q2 = 0, q3 = 0;
			invertQuaternion(q0_, q1_, q2_, q3_, ref q0, ref q1, ref q2, ref q3);

			// NOTE: put w last
			double[] arr = { q1, q2, q3, q0 };

			return arr;
		}

		double getAdaptiveGain(double alpha, double ax, double ay, double az) {
			double a_mag = Math.Sqrt(ax * ax + ay * ay + az * az);
			double error = Math.Abs(a_mag - kGravity) / kGravity;
			double factor;
			double error1 = 0.1;
			double error2 = 0.2;
			double m = 1.0 / (error1 - error2);
			double b = 1.0 - m * error1;
			if (error < error1)
				factor = 1.0;
			else if (error < error2)
				factor = m * error + b;
			else
				factor = 0.0;
			//printf("FACTOR: %f \n", factor);
			return factor * alpha;
		}

		void normalizeVector(ref double x, ref double y, ref double z) {
			double norm = Math.Sqrt(x * x + y * y + z * z);

			x /= norm;
			y /= norm;
			z /= norm;
		}

		void normalizeQuaternion(ref double q0, ref double q1, ref double q2, ref double q3) {
			double norm = Math.Sqrt(q0 * q0 + q1 * q1 + q2 * q2 + q3 * q3);
			q0 /= norm;
			q1 /= norm;
			q2 /= norm;
			q3 /= norm;
		}

		void invertQuaternion(
		  double q0, double q1, double q2, double q3,
		  ref double q0_inv, ref double q1_inv, ref double q2_inv, ref double q3_inv) {
			// Assumes quaternion is normalized.
			q0_inv = q0;
			q1_inv = -q1;
			q2_inv = -q2;
			q3_inv = -q3;
		}

		void scaleQuaternion(
		  double gain,
		  ref double dq0, ref double dq1, ref double dq2, ref double dq3) {
			if (dq0 < 0.0)//0.9
			{
				// Slerp (Spherical linear interpolation):
				double angle = Math.Acos(dq0);
				double A = Math.Sin(angle * (1.0 - gain)) / Math.Sin(angle);
				double B = Math.Sin(angle * gain) / Math.Sin(angle);
				dq0 = A + B * dq0;
				dq1 = B * dq1;
				dq2 = B * dq2;
				dq3 = B * dq3;
			} else {
				// Lerp (Linear interpolation):
				dq0 = (1.0 - gain) + gain * dq0;
				dq1 = gain * dq1;
				dq2 = gain * dq2;
				dq3 = gain * dq3;
			}

			normalizeQuaternion(ref dq0, ref dq1, ref dq2, ref dq3);
		}

		void quaternionMultiplication(
		  double p0, double p1, double p2, double p3,
		  double q0, double q1, double q2, double q3,
		  ref double r0, ref double r1, ref double r2, ref double r3) {
			// r = p q
			r0 = p0 * q0 - p1 * q1 - p2 * q2 - p3 * q3;
			r1 = p0 * q1 + p1 * q0 + p2 * q3 - p3 * q2;
			r2 = p0 * q2 - p1 * q3 + p2 * q0 + p3 * q1;
			r3 = p0 * q3 + p1 * q2 - p2 * q1 + p3 * q0;
		}

		void rotateVectorByQuaternion(
		  double x, double y, double z,
		  double q0, double q1, double q2, double q3,
		  ref double vx, ref double vy, ref double vz) {
			vx = (q0 * q0 + q1 * q1 - q2 * q2 - q3 * q3) * x + 2 * (q1 * q2 - q0 * q3) * y + 2 * (q1 * q3 + q0 * q2) * z;
			vy = 2 * (q1 * q2 + q0 * q3) * x + (q0 * q0 - q1 * q1 + q2 * q2 - q3 * q3) * y + 2 * (q2 * q3 - q0 * q1) * z;
			vz = 2 * (q1 * q3 - q0 * q2) * x + 2 * (q2 * q3 + q0 * q1) * y + (q0 * q0 - q1 * q1 - q2 * q2 + q3 * q3) * z;
		}


	}
}
