/* - Spatial full -
 * This example simply displays the spatial info if it is attached and displays the acceleration data
 * for each axis as it is changing in pretty much raw form.  It also allows for modifying the sensitivity of
 * each axis that is availabl on the attached spatial.

 * Please note that this example was designed to work with only one Phidget Spatial connected. For an
 * example showing how to use two Phidgets of the same time concurrently, please see the Servo-multi example in the Servo Examples.

 * Copyright 2019 Phidgets Inc.
 * This work is licensed under the Creative Commons Attribution 2.5 Canada License.
 * To view a copy of this license, visit http://creativecommons.org/licenses/by/2.5/ca/
 */

using OpenTK;
using OpenTK.Graphics.OpenGL;
using Phidget22;
using Phidget22.Events;
using Phidget22.ExampleUtils;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace SpatialAHRS_Example {
	public partial class Form1 : Form {
		Stream _imageStream;

		static ComplementaryAHRS ComplementaryAHRS = new ComplementaryAHRS();

		CommandLineOpen openArgs;
		Spatial openCh = null;
		formCleaner cleaner;
		private Spatial spatial;
		private TemperatureSensor temperatureSensor;
		private ErrorEventBox errorBox;
		private DataGraphBox xAccelGraphBox, yAccelGraphBox, zAccelGraphBox;
		private DataGraphBox xGyroGraphBox, yGyroGraphBox, zGyroGraphBox;
		private DataGraphBox xMagGraphBox, yMagGraphBox, zMagGraphBox;

		Bitmap spatialImage;

		double[] lastTimestamp = { 0, 0, 0 };
		bool[] lastTimestampGood = { false, false, false };
		double[] gyroHeading = { 0, 0, 0 }; //degrees

		List<double[]> compassBearingFilter = new List<double[]>();
		//int compassBearingFilterSize = 10;
		//double compassBearing = 0;
		//double ambientMagneticField = 0.57142; //Calgary

		const double ambientGravity = 1; //in G's

		Quaternion latestQuaternion = new Quaternion();
		double[] latestMag = new double[3];
		double[] latestAccel = new double[3];
		double[] latestGyro = new double[3];

		double[] lastEulerAngles = new double[3];
		double[] diagramOffsets = new double[3];

		FastEventHandler<SpatialSpatialDataEventArgs> spatialDataEventHandler;
		FastEventHandler<SpatialAlgorithmDataEventArgs> ahrsDataEventHandler;

		public Form1() {
			InitializeComponent();
			openArgs = new CommandLineOpen(this);
			cleaner = new formCleaner(this);
			algorithmCombo.DataSource = typeof(SpatialAlgorithm).ToList();
		}

		public Form1(Spatial ch) {
			InitializeComponent();
			openCh = ch;
			cleaner = new formCleaner(this);
			algorithmCombo.DataSource = typeof(SpatialAlgorithm).ToList();
		}

		//initalize the phidget device and link the event handler code
		private void Form1_Load(object sender, EventArgs e) {

			errorBox = new ErrorEventBox(this);
			xAccelGraphBox = new DataGraphBox(this, "Acceleration X");
			yAccelGraphBox = new DataGraphBox(this, "Acceleration Y");
			zAccelGraphBox = new DataGraphBox(this, "Acceleration Z");
			xGyroGraphBox = new DataGraphBox(this, "/s");
			yGyroGraphBox = new DataGraphBox(this, "/s");
			zGyroGraphBox = new DataGraphBox(this, "/s");
			xMagGraphBox = new DataGraphBox(this, "Magnetic Field Strength");
			yMagGraphBox = new DataGraphBox(this, "Magnetic Field Strength");
			zMagGraphBox = new DataGraphBox(this, "Magnetic Field Strength");

			//Tooltip for AHRS Button
			System.Windows.Forms.ToolTip ZeroARHSTooltip = new System.Windows.Forms.ToolTip();
			ZeroARHSTooltip.SetToolTip(this.zeroAHRS, "Point the cable of your spatial at the screen\nand click this to center the diagram.");

			spatial = openCh ?? openArgs.makeChannel<Spatial>();

			spatial.Attach += spatial_Attach;
			spatial.Detach += spatial_Detach;
			spatial.Error += spatial_Error;
			spatial.PropertyChange += spatial_propertychange;

			// This causes the events to get handled without invoke
			spatialDataEventHandler = new FastEventHandler<SpatialSpatialDataEventArgs>();
			spatial.SpatialData += spatialDataEventHandler.Handler;
			spatialDataEventHandler.Event += spatial_SpatialData;
			ahrsDataEventHandler = new FastEventHandler<SpatialAlgorithmDataEventArgs>();
			spatial.AlgorithmData += ahrsDataEventHandler.Handler;
			ahrsDataEventHandler.Event += Spatial_AlgorithmData;

			// Open the matching TemperatureSensor
			if (openArgs != null) {
				temperatureSensor = openArgs.makeChannel<TemperatureSensor>();
			} else {
				temperatureSensor = new TemperatureSensor();
				temperatureSensor.Channel = spatial.Channel;
				temperatureSensor.DeviceSerialNumber = spatial.DeviceSerialNumber;
				temperatureSensor.HubPort = spatial.HubPort;
				temperatureSensor.IsHubPortDevice = spatial.IsHubPortDevice;
				temperatureSensor.IsRemote = spatial.IsRemote;
				temperatureSensor.IsLocal = spatial.IsLocal;
			}

			temperatureSensor.TemperatureChange += TemperatureSensor_TemperatureChange;
			temperatureSensor.Detach += TemperatureSensor_Detach;

			try {
				spatial.Open();
			} catch (PhidgetException ex) { errorBox.addMessage("Error opening device: " + ex.Message); }
		}

		//When the application is being terminated, close the Phidget
		private void Form1_FormClosing(object sender, FormClosingEventArgs e) {

			spatial.Attach -= spatial_Attach;
			spatial.Detach -= spatial_Detach;
			spatial.Error -= spatial_Error;
			spatial.PropertyChange -= spatial_propertychange;
			spatial.SpatialData -= spatialDataEventHandler.Handler;
			spatial.AlgorithmData -= ahrsDataEventHandler.Handler;
			spatial.Close();

			temperatureSensor.TemperatureChange -= TemperatureSensor_TemperatureChange;
			temperatureSensor.Detach -= TemperatureSensor_Detach;
			temperatureSensor.Close();
		}

		private void Form1_VisibleChanged(object sender, EventArgs e) {

			if (this.Visible && spatial.Attached) {
				updateTimer.Start();
			} else {
				updateTimer.Stop();
			}
		}

		private void TemperatureSensor_Detach(object sender, DetachEventArgs e) {
			temperatureReading.Text = "Detached";
		}

		private void TemperatureSensor_TemperatureChange(object sender, TemperatureSensorTemperatureChangeEventArgs e) {
			double targetTemp = 50;
			temperatureReading.Text = e.Temperature.ToString() + "C";
			switch (((TemperatureSensor)sender).DeviceID) {
			case DeviceID.PN_MOT0109:
				targetTemp = 45;
				break;
			}
			if (e.Temperature > (targetTemp - 0.2) && e.Temperature < (targetTemp + 0.2))
				temperatureReading.ForeColor = Color.Green;
			else
				temperatureReading.ForeColor = Color.Red;
		}

		private void updateTimer_Tick(object sender, EventArgs e) {
			lock (this) {

				updateAHRS(latestQuaternion);

				glControl.Invalidate();
			}
		}

		//spatial attach event handler
		void spatial_Attach(object sender, AttachEventArgs e) {
			string imageName;

			if (this.InvokeRequired) {
				this.Invoke(new MethodInvoker(delegate () { spatial_Attach(sender, e); }));
				return;
			}

			try {
				Spatial attached = (Spatial)sender;

				//Init Graphics
				initAHRSModel();
				Assembly _assembly = Assembly.GetExecutingAssembly();
				int newWidth;
				int newHeight;
				switch (attached.DeviceSKU) {
				default:
					imageName = _assembly.GetManifestResourceNames().Where(img => img.EndsWith("1056_0_Top.png")).FirstOrDefault(); //uses the image for the 1056.  works for all models
					_imageStream = _assembly.GetManifestResourceStream(imageName);
					spatialImage = new Bitmap(_imageStream);
					newWidth = (int)((double)spatialImage.Width * ((double)compassView2.Height / (double)spatialImage.Height) * 0.6);
					newHeight = (int)((double)compassView2.Height * 0.6);
					spatialImage = new Bitmap(spatialImage, newWidth, newWidth);
					break;
				case "MOT0110":
					imageName = _assembly.GetManifestResourceNames().Where(img => img.EndsWith("Resources.MOT0110top_compass.png")).FirstOrDefault();
					_imageStream = _assembly.GetManifestResourceStream(imageName);
					spatialImage = new Bitmap(_imageStream);
					newWidth = (int)((double)spatialImage.Width * ((double)compassView2.Height / (double)spatialImage.Height) * 0.6);
					newHeight = (int)((double)compassView2.Height * 0.6);
					spatialImage = new Bitmap(spatialImage, newWidth, newWidth);
					break;
				}

				initCompassBearingGraph(compassView2);

				// End Init Graphics

				xAccelGraphBox.SetTitle(DataGraphBox.GenerateTitle("Accel X Axis", (Phidget)attached));
				yAccelGraphBox.SetTitle(DataGraphBox.GenerateTitle("Accel Y Axis", (Phidget)attached));
				zAccelGraphBox.SetTitle(DataGraphBox.GenerateTitle("Accel Z Axis", (Phidget)attached));
				xGyroGraphBox.SetTitle(DataGraphBox.GenerateTitle("Gyro X Axis", (Phidget)attached));
				yGyroGraphBox.SetTitle(DataGraphBox.GenerateTitle("Gyro Y Axis", (Phidget)attached));
				zGyroGraphBox.SetTitle(DataGraphBox.GenerateTitle("Gyro Z Axis", (Phidget)attached));
				xMagGraphBox.SetTitle(DataGraphBox.GenerateTitle("Mag X Axis", (Phidget)attached));
				yMagGraphBox.SetTitle(DataGraphBox.GenerateTitle("Mag Y Axis", (Phidget)attached));
				zMagGraphBox.SetTitle(DataGraphBox.GenerateTitle("Mag Z Axis", (Phidget)attached));

				phidgetInfoBox1.FillPhidgetInfo((Phidget)sender);

				settingsBox.Visible = true;
				ahrsCompassBox.Visible = true;
				ahrsDataBox.Visible = true;
				ahrsGraphBox.Visible = true;
				graphsBox.Visible = true;

				if (spatial.ChannelSubclass == ChannelSubclass.SpatialAHRS) {
					algorithmCombo.SelectedValue = attached.Algorithm;
					ahrsSettingsBox.Visible = true;
					fillAHRSParameterDefaults(false);
				}

				try {
					heatingEnabledBox.Checked = attached.HeatingEnabled;
					heatingEnabledPanel.Visible = true;
				} catch (PhidgetException ex) {
					if (ex.ErrorCode == Phidget22.ErrorCode.Unsupported)
						heatingEnabledPanel.Visible = false;
				}

				if (attached.MinDataInterval <= 16)
					attached.DataInterval = 16;
				else
					attached.DataInterval = attached.MinDataInterval;

				dataIntervalTrk.Unit = "ms";
				dataIntervalTrk.initializeRange(attached.MinDataInterval, attached.MaxDataInterval);
				dataIntervalTrk.Value = attached.DataInterval;

				lastTimestampGood[0] = false;
				lastTimestampGood[1] = false;
				lastTimestampGood[2] = false;
				gyroHeading[0] = 0;
				gyroHeading[1] = 0;
				gyroHeading[2] = 0;

				updateGraphDataInterval();

				updateTimer.Start();
			} catch (PhidgetException ex) {
				errorBox.addMessage("Error in attach handler: " + ex.Message);
			}

			sampleValuesBox.SelectedIndex = 0;
		}

		private void fillAHRSParameterDefaults(bool force) {
			if (angularVelBox.Text != "" && force == false)
				return;
			switch (spatial.DeviceID) {
			case DeviceID.PN_MOT0109:
			case DeviceID.PN_MOT0110:
				angularVelBox.Text = "1.0";
				angularVelDeltaBox.Text = "0.1";
				accelBox.Text = "0.05";
				magTimeBox.Text = "10";
				accelTimeBox.Text = "10";
				biasTimeBox.Text = "1.25";
				break;
			case DeviceID.PN_MOT1102:
				angularVelBox.Text = "2.0";
				angularVelDeltaBox.Text = "0.6";
				accelBox.Text = "0.1";
				magTimeBox.Text = "5";
				accelTimeBox.Text = "10";
				biasTimeBox.Text = "1.25";
				break;
			case DeviceID.PN_1044:
				angularVelBox.Text = "2.0";
				angularVelDeltaBox.Text = "0.6";
				accelBox.Text = "0.1";
				magTimeBox.Text = "5";
				accelTimeBox.Text = "10";
				biasTimeBox.Text = "1.25";
				break;
			}
		}

		//spatial detach event handler
		void spatial_Detach(object sender, DetachEventArgs e) {

			if (this.InvokeRequired) {
				this.Invoke(new MethodInvoker(delegate () { spatial_Detach(sender, e); }));
				return;
			}

			Spatial detached = (Spatial)sender;
			phidgetInfoBox1.Clear();
			cleaner.clean();

			glControl.Dispose();

			settingsBox.Visible = false;
			ahrsCompassBox.Visible = false;
			ahrsDataBox.Visible = false;
			ahrsGraphBox.Visible = false;
			ahrsSettingsBox.Visible = false;
			graphsBox.Visible = false;
			updateTimer.Stop();
			dataIntervalTrk.TrueValue = -1;
		}

		//spatial error event handler
		void spatial_Error(object sender, Phidget22.Events.ErrorEventArgs e) {

			if (this.InvokeRequired) {
				this.Invoke(new MethodInvoker(delegate () { spatial_Error(sender, e); }));
				return;
			}

			errorBox.addMessage(e.Description);
		}

		private void spatial_propertychange(object sender, PropertyChangeEventArgs e) {

			switch (e.PropertyName) {
			case "DataInterval":
				if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.INTERVAL) {
					dataIntervalTrk.Value = spatial.DataInterval;
				}
				break;
			case "DataRate":
				if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.RATE) {
					dataIntervalTrk.Value = spatial.DataRate;
				}
				break;
			case "HeatingEnabled":
				heatingEnabledBox.Checked = spatial.HeatingEnabled;
				break;
			}
		}

		void spatial_SpatialData(object sender, SpatialSpatialDataEventArgs e) {
			try {
				if (e.MagneticField[0] == (double)1e300 || e.MagneticField[1] == (double)1e300 || e.MagneticField[2] == (double)1e300)
					return;

				lock (this) {
					latestAccel = e.Acceleration;
					latestGyro = e.AngularRate;
					latestMag = e.MagneticField;
				}

				xAccelGraphBox.addData(e.Acceleration[0], spatial.DataInterval, e.Timestamp);
				yAccelGraphBox.addData(e.Acceleration[1], spatial.DataInterval, e.Timestamp);
				zAccelGraphBox.addData(e.Acceleration[2], spatial.DataInterval, e.Timestamp);
				xGyroGraphBox.addData(e.AngularRate[0], spatial.DataInterval, e.Timestamp);
				yGyroGraphBox.addData(e.AngularRate[1], spatial.DataInterval, e.Timestamp);
				zGyroGraphBox.addData(e.AngularRate[2], spatial.DataInterval, e.Timestamp);
				xMagGraphBox.addData(e.MagneticField[0], spatial.DataInterval, e.Timestamp);
				yMagGraphBox.addData(e.MagneticField[1], spatial.DataInterval, e.Timestamp);
				zMagGraphBox.addData(e.MagneticField[2], spatial.DataInterval, e.Timestamp);

				// Spatials which support AHRS in firmware
				if (spatial.ChannelSubclass == ChannelSubclass.SpatialAHRS)
					return;

				// For other spatials, run AHRS in software
				ComplementaryAHRS.update(e.Acceleration[0], e.Acceleration[1], e.Acceleration[2],
					deg2rad(e.AngularRate[0]), deg2rad(e.AngularRate[1]), deg2rad(e.AngularRate[2]),
					e.MagneticField[0], e.MagneticField[1], e.MagneticField[2], (spatial.DataInterval / 1000.0));

				double[] quaternion = ComplementaryAHRS.getOrientation();

				lock (this) {
					latestQuaternion = new Quaternion((float)quaternion[0], (float)quaternion[1], (float)quaternion[2], (float)quaternion[3]);
				}
			} catch (Exception) { }
		}
		static double deg2rad(double degrees) {
			return (Math.PI / 180) * degrees;
		}


		private void Spatial_AlgorithmData(object sender, SpatialAlgorithmDataEventArgs e) {

			lock (this) {
				latestQuaternion = new Quaternion((float)spatial.Quaternion.X, (float)spatial.Quaternion.Y, (float)spatial.Quaternion.Z, (float)spatial.Quaternion.W);
				//				latestQuaternion = new Quaternion((float)e.Quaternion[0], (float)e.Quaternion[1], (float)e.Quaternion[2], (float)e.Quaternion[3]);
			}
		}

		private void updateAHRS(Quaternion q) {

			quaternionXTxt.Text = q.X.ToString("E4");
			quaternionYTxt.Text = q.Y.ToString("E4");
			quaternionZTxt.Text = q.Z.ToString("E4");
			quaternionWTxt.Text = q.W.ToString("E4");

			SpatialEulerAngles eulerAngles;
			try {
				eulerAngles = spatial.EulerAngles;
			} catch {
				double roll = (float)Math.Atan2(2f * (q.X * q.W + q.Y * q.Z), 1 - 2f * (q.X * q.X + q.Y * q.Y));
				double pitch = (float)Math.Asin(2f * (q.W * q.Y - q.Z * q.X));
				double yaw = (float)Math.Atan2(2f * (q.W * q.Z + q.X * q.Y), 1 - 2f * (q.Y * q.Y + q.Z * q.Z));

				// Convert radians to degrees
				yaw = yaw * 180 / Math.PI;
				pitch = pitch * 180 / Math.PI;
				roll = roll * 180 / Math.PI;

				eulerAngles = new SpatialEulerAngles() { Roll = roll, Pitch = pitch, Heading = yaw };
			}
			lastEulerAngles[0] = eulerAngles.Roll;
			lastEulerAngles[1] = eulerAngles.Pitch;
			lastEulerAngles[2] = eulerAngles.Heading;

			//x_IMU_API.QuaternionData quaternionD = new x_IMU_API.QuaternionData(Array.ConvertAll(e.Quaternion, x => (float)x));
			//float[] angles = quaternionD.ConvertToConjugate().ConvertToEulerAngles();
			eulerRoll.Text = eulerAngles.Roll.ToString("F3") + "";
			eulerPitch.Text = eulerAngles.Pitch.ToString("F3") + "";
			eulerHeading.Text = eulerAngles.Heading.ToString("F2") + "";

			drawCompassBearingGraph(compassView2, eulerAngles.Heading);

			q.Conjugate();
			RotationMatrix = Matrix3.CreateFromQuaternion(q);
		}

		private void updateGraphDataInterval() {
			try {
				xAccelGraphBox.DataInterval = spatial.DataInterval;
				yAccelGraphBox.DataInterval = spatial.DataInterval;
				zAccelGraphBox.DataInterval = spatial.DataInterval;
				xGyroGraphBox.DataInterval = spatial.DataInterval;
				yGyroGraphBox.DataInterval = spatial.DataInterval;
				zGyroGraphBox.DataInterval = spatial.DataInterval;
				xMagGraphBox.DataInterval = spatial.DataInterval;
				yMagGraphBox.DataInterval = spatial.DataInterval;
				zMagGraphBox.DataInterval = spatial.DataInterval;
			} catch (PhidgetException ex) { errorBox.addMessage("Error updating graph data interval: " + ex.Message); }
		}
		private void zeroAlgBtn_Click(object sender, EventArgs e) {
			try {
				spatial.ZeroAlgorithm();
			} catch (Exception ex) {
				errorBox.addMessage("Error zeroing AHRS: " + ex.Message);
			}
		}

		private void algorithmCombo_SelectedIndexChanged(object sender, EventArgs e) {

			if (spatial == null)
				return;
			try {
				spatial.Algorithm = (SpatialAlgorithm)Enum.Parse(typeof(SpatialAlgorithm), algorithmCombo.SelectedValue.ToString());
				if (spatial.Algorithm == SpatialAlgorithm.AHRS)
					ahrsCompassBox.Text = "AHRS Compass";
				else if (spatial.Algorithm == SpatialAlgorithm.IMU)
					ahrsCompassBox.Text = "Heading";
			} catch (PhidgetException ex) {
				errorBox.addMessage("Error setting algorithm: " + ex.Message);
			}
		}

		private void dataIntervalTrk_KeyOverride(object sender, KeyEventArgs e) {
			e.SuppressKeyPress = true;
			dataIntervalTrk.keyOverride(sender, e);
		}

		private void slider_ValueChangeEvent(object sender, EventArgs e) {
			if (((PhidgetsControlLibrary.CustomScroll)sender).isFixedRange) //datainterval control
			{
				if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.INTERVAL) {
					try {
						double temp = dataIntervalTrk.Value;
						spatial.DataInterval = (int)temp;
						if (spatial.DataInterval < (int)temp) {
							while (spatial.DataInterval != (int)temp) {
								temp++;
								spatial.DataInterval = (int)temp;
							}
						} else {
							while (spatial.DataInterval != (int)temp) {
								temp--;
								spatial.DataInterval = (int)temp;
							}
						}

						dataIntervalTrk.TrueValue = spatial.DataInterval;
						dataIntervalTrk.Value = spatial.DataInterval;



						updateGraphDataInterval();
					} catch (Exception ex) { errorBox.addMessage("Error setting data interval: " + ex.Message); }
				} else if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.RATE) {
					try {
						spatial.DataRate = dataIntervalTrk.Value;
						updateGraphDataInterval();
					} catch (Exception ex) { errorBox.addMessage("Error setting data rate: " + ex.Message); }
				}
			}
		}
		private void dataIntervalTrk_ModeSwitch(object sender, EventArgs e) {
			dataIntervalTrk.SwitchModes();
			if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.RATE) {
				try {
					dataIntervalTrk.isIntValue = false;
					dataIntervalTrk.Unit = "Hz";
					dataIntervalTrk.trkText = spatial.DataRate.ToString() + "Hz";
					dataIntervalTrk.Value = spatial.DataRate;
				} catch (Exception ex) { errorBox.addMessage("Error setting data rate: " + ex.Message); }

			} else if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.INTERVAL) {
				try {
					dataIntervalTrk.isIntValue = true;
					if (spatial.DeviceID != DeviceID.PN_1046)
						dataIntervalTrk.Unit = "ms";
					dataIntervalTrk.trkText = spatial.DataInterval.ToString() + "ms";
					dataIntervalTrk.Value = spatial.DataInterval;
				} catch (Exception ex) { errorBox.addMessage("Error setting data interval: " + ex.Message); }
			}
			dataIntervalTrk.TrueValue = dataIntervalTrk.Value;
		}

		#region Graphs

		private BufferedGraphicsContext compassGraphicsContext;
		private BufferedGraphics compassGraphicsBuffer;
		private Bitmap compassDrawingSurface;

		private RectangleF compasscircleRectangle;
		private Rectangle compassboundsRectangle;
		private Pen compassDotPen, compassCirclePen, compassTickPen, compassTickPenBig;
		private float compassCircleDiameter, compassCircleRadius;

		private void initCompassBearingGraph(Panel compassView) {

			//initialize our graphical representation of the compasseration data
			compassDrawingSurface = new Bitmap(compassView.ClientRectangle.Width, compassView.ClientRectangle.Height, System.Drawing.Imaging.PixelFormat.Format24bppRgb);

			compassboundsRectangle = new Rectangle(0, 0, compassDrawingSurface.Width, compassDrawingSurface.Height);
			compassboundsRectangle.Inflate(-1, -1);

			float xCenter = (float)(compassboundsRectangle.Width / 2.0);
			float yCenter = (float)(compassboundsRectangle.Height / 2.0);

			compassDotPen = new Pen(Color.Red, 4);
			compassCirclePen = new Pen(Color.Black, 2);
			compassTickPen = new Pen(Color.DarkBlue, 1);
			compassTickPenBig = new Pen(Color.DarkBlue, 2);

			compassCircleDiameter = (float)(compassDrawingSurface.Width / 1.35);
			compassCircleRadius = (float)(compassCircleDiameter / 2.0);

			compasscircleRectangle = new RectangleF(xCenter - compassCircleRadius, yCenter - compassCircleRadius, compassCircleDiameter, compassCircleDiameter);
		}

		private void drawCompassBearingGraph(Panel compassView, double compassBearing) {
			compassGraphicsContext = new BufferedGraphicsContext();
			compassGraphicsBuffer = compassGraphicsContext.Allocate(Graphics.FromImage(compassDrawingSurface), compassboundsRectangle);
			compassGraphicsBuffer.Graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
			compassGraphicsBuffer.Graphics.Clear(SystemColors.Control);

			float xCenter = (float)(compassboundsRectangle.Width / 2.0);
			float yCenter = (float)(compassboundsRectangle.Height / 2.0);

			compassGraphicsBuffer.Graphics.DrawEllipse(compassCirclePen, compasscircleRectangle);

			Point pt = new Point((int)(xCenter - spatialImage.Width / 2), (int)(yCenter - spatialImage.Height / 2));
			compassGraphicsBuffer.Graphics.DrawImageUnscaled(rotateImage(spatialImage, (float)compassBearing), pt);

			string topText, rightText, bottomText, leftText;
			float topX, topY, rightX, rightY, bottomX, bottomY, leftX, leftY;
			switch (spatial.Algorithm) {
			case SpatialAlgorithm.AHRS:
				topText = "N";
				topX = xCenter - 10;
				topY = 0;

				rightText = "E";
				rightX = xCenter * 2 - 20;
				rightY = yCenter - 12;

				bottomText = "S";
				bottomX = xCenter - 10;
				bottomY = yCenter * 2 - 20;

				leftText = "W";
				leftX = 0;
				leftY = yCenter - 12;
				break;
			case SpatialAlgorithm.IMU:
				topText = "0";
				topX = xCenter - 10;
				topY = 0;

				rightText = "90";
				rightX = xCenter * 2 - 35;
				rightY = yCenter - 12;

				bottomText = "180";
				bottomX = xCenter - 20;
				bottomY = yCenter * 2 - 20;

				leftText = "-90";
				leftX = -10;
				leftY = yCenter - 12;
				break;
			default:
				topText = "";
				topX = xCenter - 10;
				topY = 0;

				rightText = "";
				rightX = xCenter * 2 - 20;
				rightY = yCenter - 12;

				bottomText = "";
				bottomX = xCenter - 10;
				bottomY = yCenter * 2 - 20;

				leftText = "";
				leftX = 0;
				leftY = yCenter - 12;
				break;
			}

			compassGraphicsBuffer.Graphics.DrawString(topText, new Font("Arial", 16), new SolidBrush(Color.Black), topX, topY);
			compassGraphicsBuffer.Graphics.DrawString(rightText, new Font("Arial", 16), new SolidBrush(Color.Black), rightX, rightY);
			compassGraphicsBuffer.Graphics.DrawString(bottomText, new Font("Arial", 16), new SolidBrush(Color.Black), bottomX, bottomY);
			compassGraphicsBuffer.Graphics.DrawString(leftText, new Font("Arial", 16), new SolidBrush(Color.Black), leftX, leftY);

			//Ticks around the compass circle
			for (int i = 0; i < 360; i += 10) {
				Pen p = compassTickPen;
				int tickSize = 4;
				if (i == 0 || i == 90 || i == 180 || i == 270) {
					p = compassTickPenBig;
					tickSize = 6;
				}
				compassGraphicsBuffer.Graphics.DrawLine(p,
				(float)((compassCircleRadius + tickSize) * Math.Cos(i * (Math.PI / 180.0)) + xCenter),
				(float)(-(compassCircleRadius + tickSize) * Math.Sin(i * (Math.PI / 180.0)) + yCenter),
				(float)((compassCircleRadius - tickSize) * Math.Cos(i * (Math.PI / 180.0)) + xCenter),
				(float)(-(compassCircleRadius - tickSize) * Math.Sin(i * (Math.PI / 180.0)) + yCenter));
			}

			//Marker on the compass circle
			compassGraphicsBuffer.Graphics.DrawEllipse(
				compassDotPen,
				(float)(compassCircleRadius * Math.Cos((-compassBearing + 90) * (Math.PI / 180.0)) + xCenter) - 2,
				(float)(-compassCircleRadius * Math.Sin((-compassBearing + 90) * (Math.PI / 180.0)) + yCenter) - 2,
				4,
				4
			);

			compassGraphicsBuffer.Render(compassView.CreateGraphics());

			compassGraphicsBuffer.Dispose();

			compassGraphicsContext.Dispose();
		}
		#endregion

		#region AHRS Model

		private string[] imageFiles;

		private void zeroAHRS_Click(object sender, EventArgs e) {
			diagramOffsets[0] = lastEulerAngles[0];
			diagramOffsets[1] = lastEulerAngles[1];
			diagramOffsets[2] = lastEulerAngles[2];
		}

		private async void heatingEnabledBox_CheckedChanged(object sender, EventArgs e) {
			try {
				heatingEnabledBox.Enabled = false;
				spatial.HeatingEnabled = heatingEnabledBox.Checked;
				if (heatingEnabledBox.Checked) {
					try {
						await Task.Run(() => { temperatureSensor.Open(1000); }); //short timeout so that if the temp is already open on another channel we just ignore it
						tempReleaseTimer.Start(); //start checking the temperature.
					} catch (PhidgetException ex) {
						errorBox.addMessage("Error opening temperature channel: " + ex.Message);
					}
				} else if (!heatingEnabledBox.Checked) { //close the channel when heating isn't important to improve speed
					temperatureSensor.Close();
					tempReleaseTimer.Stop();
					temperatureReading.Text = "Unknown";
					temperatureReading.ForeColor = Color.Black;
				}
			} catch (PhidgetException ex) {
				errorBox.addMessage("Error setting heating enabled: " + ex.Message);
			} finally {
				heatingEnabledBox.Enabled = true;
			}
		}
		private void tempReleaseTimer_Tick(object sender, EventArgs e) {
			try {
				if (temperatureSensor.Temperature >= checkSetPoint(spatial.DeviceID)) {
					//we have reached the setpoint, we can close the temperature sensor channel now and just display a stable temp message.  this allows for
					//higher data rates to be reached since less data is coming back
					temperatureSensor.Close();
					temperatureReading.ForeColor = Color.Green;
					temperatureReading.Text = "Temp Stabilized";
					tempReleaseTimer.Stop();
				}
			} catch (PhidgetException) { }
		}
		private double checkSetPoint(Phidget22.DeviceID id) {
			switch (id) {
			case DeviceID.PN_MOT0109:
				return 45.0;
			case DeviceID.PN_MOT0110:
				return 50.0;
			default:
				return -1;
			}
		}

		private uint[] textures;

		private float halfXdimension, halfYdimension, halfZdimension;

		private float[] transformationMatrix;
		private float CameraDistance;

		private OpenTK.GLControl glControl;

		private void button2_Click(object sender, EventArgs e) {
			spatial.ZeroGyroscope();
		}

		private void SetAHRSParamsButton_Click(object sender, EventArgs e) {
			spatial.SetAHRSParameters(Convert.ToDouble(angularVelBox.Text), Convert.ToDouble(angularVelDeltaBox.Text), Convert.ToDouble(accelBox.Text),
									Convert.ToDouble(magTimeBox.Text), Convert.ToDouble(accelTimeBox.Text), Convert.ToDouble(biasTimeBox.Text));
		}

		private void button1_Click(object sender, EventArgs e) {
			spatial.Precision = SpatialPrecision.Hybrid;
		}

		private void xAccelGraphBtn_Click(object sender, EventArgs e) {
			if (!xAccelGraphBox.Visible)
				xAccelGraphBox.Show();
		}

		private void yAccelGraphBtn_Click(object sender, EventArgs e) {
			if (!yAccelGraphBox.Visible)
				yAccelGraphBox.Show();
		}

		private void zAccelGraphBtn_Click(object sender, EventArgs e) {
			if (!zAccelGraphBox.Visible)
				zAccelGraphBox.Show();
		}

		private void xGyroGraphBtn_Click(object sender, EventArgs e) {
			if (!xGyroGraphBox.Visible)
				xGyroGraphBox.Show();
		}

		private void yGyroGraphBtn_Click(object sender, EventArgs e) {
			if (!yGyroGraphBox.Visible)
				yGyroGraphBox.Show();
		}

		private void zGyroGraphBtn_Click(object sender, EventArgs e) {
			if (!zGyroGraphBox.Visible)
				zGyroGraphBox.Show();
		}

		private void xMagGraphBtn_Click(object sender, EventArgs e) {
			if (!xMagGraphBox.Visible)
				xMagGraphBox.Show();
		}

		private void yMagGraphBtn_Click(object sender, EventArgs e) {
			if (!yMagGraphBox.Visible)
				yMagGraphBox.Show();
		}

		private void zMagGraphBtn_Click(object sender, EventArgs e) {
			if (!zMagGraphBox.Visible)
				zMagGraphBox.Show();
		}

		private void button3_Click(object sender, EventArgs e) {
			spatial.Precision = SpatialPrecision.High;
		}

		private void button4_Click(object sender, EventArgs e) {
			spatial.Precision = SpatialPrecision.Low;
		}

		private void sampleValuesBox_SelectedIndexChanged(object sender, EventArgs e) {
			switch (sampleValuesBox.SelectedIndex) {
			case 0: //General Purpose
				fillAHRSParameterDefaults(true);
				break;
			case 1: //No Gyro
				angularVelBox.Text = "0.001";
				angularVelDeltaBox.Text = "0.5";
				accelBox.Text = "2";
				magTimeBox.Text = "0";
				accelTimeBox.Text = "0";
				biasTimeBox.Text = "1.25";
				break;
			case 2: //All Gyro
				angularVelBox.Text = "0.00";
				angularVelDeltaBox.Text = "0.00";
				accelBox.Text = "0.00";
				magTimeBox.Text = "1000";
				accelTimeBox.Text = "1000";
				biasTimeBox.Text = "1000";
				break;
			case 3: //Zeroed
				switch (spatial.DeviceID) {
				case DeviceID.PN_MOT0109:
				case DeviceID.PN_MOT0110:
					angularVelBox.Text = "0.5";
					angularVelDeltaBox.Text = "0.1";
					accelBox.Text = "0.05";
					magTimeBox.Text = "120";
					accelTimeBox.Text = "120";
					biasTimeBox.Text = "1.25";
					break;
				case DeviceID.PN_MOT1102:
					angularVelBox.Text = "1.0";
					angularVelDeltaBox.Text = "0.6";
					accelBox.Text = "0.1";
					magTimeBox.Text = "5";
					accelTimeBox.Text = "30";
					biasTimeBox.Text = "1.25";
					break;
				case DeviceID.PN_1044:
					angularVelBox.Text = "2.0";
					angularVelDeltaBox.Text = "0.6";
					accelBox.Text = "0.1";
					magTimeBox.Text = "5";
					accelTimeBox.Text = "30";
					biasTimeBox.Text = "1.25";
					break;
				}
				break;
			case 4: //High Vibrations
				angularVelBox.Text = "5.0";
				angularVelDeltaBox.Text = "2.5";
				accelBox.Text = "0.2";
				magTimeBox.Text = "5";
				accelTimeBox.Text = "20";
				biasTimeBox.Text = "5";
				break;
			}
			SetAHRSParamsButton_Click(SetAHRSParamsButton, e);
		}
		private void initAHRSModel() {
			String sku;
			// Nicer, but slower
			glControl = new OpenTK.GLControl(new OpenTK.Graphics.GraphicsMode(32, 24, 0, 16));
			//glControl = new OpenTK.GLControl();

			ahrsGraphBox.Controls.Add(this.glControl);

			glControl.Dock = System.Windows.Forms.DockStyle.Fill;
			glControl.VSync = false;
			glControl.Load += new System.EventHandler(this.simpleOpenGlControl_Load);
			glControl.SizeChanged += new System.EventHandler(this.simpleOpenGlControl_SizeChanged);
			glControl.Paint += new System.Windows.Forms.PaintEventHandler(this.simpleOpenGlControl_Paint);

			transformationMatrix = new float[] {1.0f, 0.0f, 0.0f, 0.0f,
												0.0f, 1.0f, 0.0f, 0.0f,
												0.0f, 0.0f, 1.0f, 0.0f,
												0.0f, 0.0f, 0.0f, 1.0f};

			try {
				sku = spatial.DeviceSKU;
			} catch {
				sku = "";
			}
			switch (sku) {
			case "MOT0110":
				imageFiles = new string[] { "Resources.MOT0110side.png", "Resources.MOT0110side.png",
						"Resources.MOT0110side.png", "Resources.MOT0110front.png",
						"Resources.MOT0110bottom.png", "Resources.MOT0110top.png" };
				break;
			default:
				imageFiles = new string[] { "Resources.left.png", "Resources.right.png",
						"Resources.back.png", "Resources.front.png",
						"Resources.bottom.png", "Resources.top.png" };
				break;
			}

			float[] dimensions = new float[] { 6, 6, 2 };
			halfXdimension = dimensions[0] / 2;
			halfYdimension = dimensions[1] / 2;
			halfZdimension = dimensions[2] / 2;
			CameraDistance = 50;
		}

		private Matrix3 RotationMatrix {
			set {
				transformationMatrix[0] = value.M11; transformationMatrix[4] = value.M12; transformationMatrix[8] = value.M13;
				transformationMatrix[1] = value.M21; transformationMatrix[5] = value.M22; transformationMatrix[9] = value.M23;
				transformationMatrix[2] = value.M31; transformationMatrix[6] = value.M32; transformationMatrix[10] = value.M33;
			}
		}

		private uint[] LoadTextureFromImage(string[] filesNames) {
			int numOfPic = filesNames.Length;
			uint[] texture = new uint[numOfPic];
			Bitmap[] bitmap = new Bitmap[numOfPic];
			BitmapData[] bitmapdata = new BitmapData[numOfPic];
			for (int im = 0; im < numOfPic; im++) {
				Assembly _assembly = Assembly.GetExecutingAssembly();
				var _resources = _assembly.GetManifestResourceNames();
				string imageName = _resources.Where(img => img.EndsWith(filesNames[im])).FirstOrDefault();
				Stream _imageStream = _assembly.GetManifestResourceStream(imageName);
				bitmap[im] = new Bitmap(_imageStream);
			}
			GL.GenTextures(numOfPic, texture);
			for (int i = 0; i < numOfPic; i++) {

				bitmap[i].RotateFlip(RotateFlipType.RotateNoneFlipY);
				Rectangle rect = new Rectangle(0, 0, bitmap[i].Width, bitmap[i].Height);
				bitmapdata[i] = bitmap[i].LockBits(rect, ImageLockMode.ReadOnly, System.Drawing.Imaging.PixelFormat.Format32bppArgb);

				GL.BindTexture(TextureTarget.Texture2D, texture[i]);

				GL.TexEnv(TextureEnvTarget.TextureEnv, TextureEnvParameter.TextureEnvMode, (int)TextureEnvMode.Modulate);
				GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)TextureMagFilter.Nearest);
				GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Nearest);

				GL.TexImage2D(TextureTarget.Texture2D, 0, PixelInternalFormat.Rgba, bitmap[i].Width, bitmap[i].Height, 0,
					OpenTK.Graphics.OpenGL.PixelFormat.Bgra, PixelType.UnsignedByte, IntPtr.Zero);

				GL.TexSubImage2D(TextureTarget.Texture2D, 0, 0, 0, bitmap[i].Width, bitmap[i].Height, OpenTK.Graphics.OpenGL.PixelFormat.Bgra,
					PixelType.UnsignedByte, bitmapdata[i].Scan0);

				bitmap[i].UnlockBits(bitmapdata[i]);
				bitmap[i].Dispose();
				GL.BindTexture(TextureTarget.Texture2D, 0);
			}
			return texture;
		}

		private void simpleOpenGlControl_Load(object sender, EventArgs e) {

			GL.ClearColor(this.BackColor);

			glControl.SwapBuffers();
			simpleOpenGlControl_SizeChanged(sender, e);

			GL.ShadeModel(ShadingModel.Smooth);
			GL.Enable(EnableCap.LineSmooth);
			GL.Enable(EnableCap.Texture2D);
			GL.Enable(EnableCap.Normalize);
			GL.Enable(EnableCap.ColorMaterial);
			GL.Enable(EnableCap.DepthTest);
			GL.Enable(EnableCap.Blend);
			//GL.Enable(EnableCap.Lighting);
			//GL.Enable(EnableCap.Light0);
			GL.Hint(HintTarget.PolygonSmoothHint, HintMode.Nicest);
			textures = LoadTextureFromImage(imageFiles);
		}

		private void simpleOpenGlControl_Paint(object sender, PaintEventArgs e) {

			glControl.MakeCurrent();

			GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);
			GL.LoadIdentity();
			GL.PolygonMode(MaterialFace.Front, PolygonMode.Fill);

			// Set camera view and distance
			GL.Translate(0, 0, -1.0f * CameraDistance);

			// Show a 'front' view - rotate the 1044 coordinate system into the OpenGL systm
			GL.Rotate(-90, 1, 0, 0);
			GL.Rotate(180, 0, 1, 0);
			GL.Rotate(180 - diagramOffsets[2], 0, 0, 1);

			GL.PushMatrix();
			GL.MultMatrix(transformationMatrix);

			// +'ve x face
			GL.BindTexture(TextureTarget.Texture2D, textures[0]);
			GL.Begin(PrimitiveType.Quads);
			GL.Normal3(1, 0, 0); GL.TexCoord2(0, 0); GL.Vertex3(halfXdimension, -halfYdimension, -halfZdimension);
			GL.Normal3(1, 0, 0); GL.TexCoord2(0, 1); GL.Vertex3(halfXdimension, -halfYdimension, halfZdimension);
			GL.Normal3(1, 0, 0); GL.TexCoord2(1, 1); GL.Vertex3(halfXdimension, halfYdimension, halfZdimension);
			GL.Normal3(1, 0, 0); GL.TexCoord2(1, 0); GL.Vertex3(halfXdimension, halfYdimension, -halfZdimension);
			GL.End();

			// -'ve x face
			GL.BindTexture(TextureTarget.Texture2D, textures[1]);
			GL.Begin(PrimitiveType.Quads);
			GL.Normal3(-1, 0, 0); GL.TexCoord2(1, 0); GL.Vertex3(-halfXdimension, -halfYdimension, -halfZdimension);
			GL.Normal3(-1, 0, 0); GL.TexCoord2(1, 1); GL.Vertex3(-halfXdimension, -halfYdimension, halfZdimension);
			GL.Normal3(-1, 0, 0); GL.TexCoord2(0, 1); GL.Vertex3(-halfXdimension, halfYdimension, halfZdimension);
			GL.Normal3(-1, 0, 0); GL.TexCoord2(0, 0); GL.Vertex3(-halfXdimension, halfYdimension, -halfZdimension);
			GL.End();

			// +'ve y face
			GL.BindTexture(TextureTarget.Texture2D, textures[2]);
			GL.Begin(PrimitiveType.Quads);
			GL.Normal3(0, 1, 0); GL.TexCoord2(1, 0); GL.Vertex3(-halfXdimension, halfYdimension, -halfZdimension);
			GL.Normal3(0, 1, 0); GL.TexCoord2(1, 1); GL.Vertex3(-halfXdimension, halfYdimension, halfZdimension);
			GL.Normal3(0, 1, 0); GL.TexCoord2(0, 1); GL.Vertex3(halfXdimension, halfYdimension, halfZdimension);
			GL.Normal3(0, 1, 0); GL.TexCoord2(0, 0); GL.Vertex3(halfXdimension, halfYdimension, -halfZdimension);
			GL.End();

			// -'ve y face
			GL.BindTexture(TextureTarget.Texture2D, textures[3]);
			GL.Begin(PrimitiveType.Quads);
			GL.Normal3(0, -1, 0); GL.TexCoord2(0, 0); GL.Vertex3(-halfXdimension, -halfYdimension, -halfZdimension);
			GL.Normal3(0, -1, 0); GL.TexCoord2(0, 1); GL.Vertex3(-halfXdimension, -halfYdimension, halfZdimension);
			GL.Normal3(0, -1, 0); GL.TexCoord2(1, 1); GL.Vertex3(halfXdimension, -halfYdimension, halfZdimension);
			GL.Normal3(0, -1, 0); GL.TexCoord2(1, 0); GL.Vertex3(halfXdimension, -halfYdimension, -halfZdimension);
			GL.End();

			// +'ve z face
			GL.BindTexture(TextureTarget.Texture2D, textures[4]);
			GL.Begin(PrimitiveType.Quads);
			GL.Normal3(0, 0, 1); GL.TexCoord2(0, 0); GL.Vertex3(-halfXdimension, -halfYdimension, halfZdimension);
			GL.Normal3(0, 0, 1); GL.TexCoord2(1, 0); GL.Vertex3(halfXdimension, -halfYdimension, halfZdimension);
			GL.Normal3(0, 0, 1); GL.TexCoord2(1, 1); GL.Vertex3(halfXdimension, halfYdimension, halfZdimension);
			GL.Normal3(0, 0, 1); GL.TexCoord2(0, 1); GL.Vertex3(-halfXdimension, halfYdimension, halfZdimension);
			GL.End();

			// -'ve z face
			GL.BindTexture(TextureTarget.Texture2D, textures[5]);
			GL.Begin(PrimitiveType.Quads);
			GL.Normal3(0, 0, -1); GL.TexCoord2(0, 1); GL.Vertex3(-halfXdimension, -halfYdimension, -halfZdimension);
			GL.Normal3(0, 0, -1); GL.TexCoord2(1, 1); GL.Vertex3(halfXdimension, -halfYdimension, -halfZdimension);
			GL.Normal3(0, 0, -1); GL.TexCoord2(1, 0); GL.Vertex3(halfXdimension, halfYdimension, -halfZdimension);
			GL.Normal3(0, 0, -1); GL.TexCoord2(0, 0); GL.Vertex3(-halfXdimension, halfYdimension, -halfZdimension);
			GL.End();

			GL.PopMatrix();
			GL.Flush();

			glControl.SwapBuffers();
		}

		private void simpleOpenGlControl_SizeChanged(object sender, EventArgs e) {

			int height = glControl.Size.Height;
			int width = glControl.Size.Width;
			GL.Viewport(0, 0, width, height);
			GL.PushMatrix();

			float lAngleView = (float)(10 * Math.PI / 180f);
			float lAspectRatio = (float)width / (float)height;
			float lDistanceToNearClipPlane = 1f;
			float lDistanceToFarClipPlane = 250f;

			Matrix4 perspectiveMatrix = Matrix4.CreatePerspectiveFieldOfView(lAngleView, lAspectRatio, lDistanceToNearClipPlane, lDistanceToFarClipPlane);

			GL.MatrixMode(MatrixMode.Projection);
			GL.LoadIdentity();
			GL.LoadMatrix(ref perspectiveMatrix);

			GL.MatrixMode(MatrixMode.Modelview);
			GL.PopMatrix();
		}

		private Bitmap rotateImage(Bitmap b, float angle) {
			//create a new empty bitmap to hold rotated image
			Bitmap returnBitmap = new Bitmap(b.Width, b.Height);
			//make a graphics object from the empty bitmap
			Graphics g = Graphics.FromImage(returnBitmap);
			//move rotation point to center of image
			g.TranslateTransform((float)b.Width / 2, (float)b.Height / 2);
			//rotate
			g.RotateTransform(angle);
			//move image back
			g.TranslateTransform(-(float)b.Width / 2, -(float)b.Height / 2);
			//draw passed in image onto graphics object
			g.DrawImage(b, new Point(0, 0));
			return returnBitmap;
		}

		#endregion
	}
}